///////////////////////////////////////////////////////////////////////////////
//  Filename: $RCSfile: Hungarian2.java,v $
//  Purpose:  Solving the Minimum Assignment Problem using the
//            Hungarian Method.
//  Language: Java
//  Compiler: JDK 1.4
//  Authors:  Joerg K. Wegner
//  Version:  $Revision: 1.4 $
//            $Date: 2004/09/13 06:52:25 $
//            $Author: wegnerj $
//  Original authors:  Cyrill Stachnis
//
//  Copyright (c) Dept. Computer Architecture, University of Tuebingen, Germany
//
// Parts of the used code was originally provided by the 
// "Stanford GraphGase", but I made changes to this code.
// As asked by  the copyright node of the "Stanford GraphGase", 
// I hereby proclaim that this file are *NOT* part of the
// "Stanford GraphGase" distrubition!
// 
// This file is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.
///////////////////////////////////////////////////////////////////////////////
package Alignment;



/**
 * Solving the Minimum Assignment Problem using the Hungarian Method.
 *
 * @author     wegnerj
 */
public class Hungarian2
{

    private static final int HUNGARIAN_NOT_ASSIGNED = 0;
    private static final int HUNGARIAN_ASSIGNED = 1;

    //~ Methods ////////////////////////////////////////////////////////////////

    public double init(HungarianProblem p, double[][] costMatrix,
        boolean maximize)
    {
        int i;
        int j;
        int orgCols;
        int orgRows;
        double maxCost = 0;

        orgCols = costMatrix[0].length;
        orgRows = costMatrix.length;

        // is the number of cols  not equal to number of rows ? 
        // if yes, expand with 0-cols / 0-cols
        int rows = Math.max(orgCols, orgRows);
        int cols = rows;

        p.numRows = rows;
        p.numCols = cols;

        //logger.info("rows="+rows+" cols="+cols);
        p.cost = new double[rows][cols];
        p.assignment = new int[rows][cols];
        //NEU: FOBER
        p.zuordnung = new int[2][costMatrix.length];
        //ENDE
        for (i = 0; i < p.numRows; i++)
        {
            for (j = 0; j < p.numCols; j++)
            {
                p.cost[i][j] = ((i < orgRows) && (j < orgCols))
                    ? costMatrix[i][j] : 0;

                p.assignment[i][j] = 0;

                if (maxCost < p.cost[i][j])
                {
                    maxCost = p.cost[i][j];
                }
            }
        }

        //System.out.println("maxcost "+maxCost);
        if (maximize)
        {
            for (i = 0; i < p.numRows; i++)
            {
                for (j = 0; j < p.numCols; j++)
                {
                    p.cost[i][j] = maxCost - p.cost[i][j];
                }
            }
        }

        return maxCost;
    }

    public double solve(HungarianProblem p)
    {
        int i;
        int j;
        int m;
        int n;
        int k;
        int l;
        int t;
        int q;
        int unmatched;
        double s;
        double cost;
        int[] colMate;
        int[] rowMate;
        int[] parentRow;
        int[] unchosenRow;
        double[] rowDec;
        double[] colInc;
        double[] slack;
        int[] slackRow;

        cost = 0;
        m = p.numRows;
        n = p.numCols;

        colMate = new int[p.numRows];
        unchosenRow = new int[p.numRows];
        rowDec = new double[p.numRows];
        slackRow = new int[p.numRows];

        rowMate = new int[p.numCols];
        parentRow = new int[p.numCols];
        colInc = new double[p.numCols];
        slack = new double[p.numCols];

        for (i = 0; i < p.numRows; i++)
        {
            colMate[i] = 0;
            unchosenRow[i] = 0;
            rowDec[i] = 0;
            slackRow[i] = 0;
        }

        for (j = 0; j < p.numCols; j++)
        {
            rowMate[j] = 0;
            parentRow[j] = 0;
            colInc[j] = 0;
            slack[j] = 0;
        }

        for (i = 0; i < p.numRows; ++i)
        {
            for (j = 0; j < p.numCols; ++j)
            {
                p.assignment[i][j] = HUNGARIAN_NOT_ASSIGNED;
            }
        }

        // Begin subtract column minima in order to start with lots of zeroes 12
        for (l = 0; l < n; l++)
        {
            s = p.cost[0][l];

            for (k = 1; k < m; k++)
            {
                if (p.cost[k][l] < s)
                {
                    s = p.cost[k][l];
                }
            }

            cost += s;

            if (s != 0)
            {
                for (k = 0; k < m; k++)
                {
                    p.cost[k][l] -= s;
                }
            }
        }

        // End subtract column minima in order to start with lots of zeroes 12
        // Begin initial state 16
        t = 0;

        for (l = 0; l < n; l++)
        {
            rowMate[l] = -1;
            parentRow[l] = -1;
            colInc[l] = 0;
            slack[l] = Double.POSITIVE_INFINITY;
        }

        boolean rowDone = false;

        for (k = 0; k < m; k++)
        {
            s = p.cost[k][0];

            for (l = 1; l < n; l++)
            {
                if (p.cost[k][l] < s)
                {
                    s = p.cost[k][l];
                }
            }

            rowDec[k] = s;
            rowDone = false;

            for (l = 0; l < n; l++)
            {
                if ((s == p.cost[k][l]) && (rowMate[l] < 0))
                {
                    colMate[k] = l;
                    rowMate[l] = k;
                    rowDone = true;

                    break; // for-loop
                }
            }

            if (rowDone == false)
            {
                colMate[k] = -1;
                unchosenRow[t++] = k;
            }
        }

        // End initial state 16
        // Begin Hungarian algorithm 18
        boolean done = false;

        if (t == 0)
        {
            done = true;
        }

        if (done == false)
        {
            unmatched = t;

            boolean breakthru = false;

            while (true)
            {
                breakthru = false;
                q = 0;

                while (breakthru == false)
                {
                    while ((q < t) && (breakthru == false))
                    {
                        k = unchosenRow[q];
                        s = rowDec[k];

                        for (l = 0; l < n; l++)
                        {
                            if (slack[l] != 0)
                            {
                                double del;
                                del = p.cost[k][l] - s + colInc[l];

                                if (del < slack[l])
                                {
                                    if (del == 0)
                                    {
                                        if (rowMate[l] < 0)
                                        {
                                            breakthru = true;
                                            break; // for-loop
                                        }

                                        slack[l] = 0;
                                        parentRow[l] = k;
                                        unchosenRow[t++] = rowMate[l];
                                    }
                                    else
                                    {
                                        slack[l] = del;
                                        slackRow[l] = k;
                                    }
                                }
                            }
                        }

                        // End explore node q of the forest 19
                        if (breakthru == false)
                        {
                            q++;
                        }
                    }

                    if (breakthru == false)
                    {
                        // Begin introduce a new zero into the matrix 21
                        s = Double.POSITIVE_INFINITY;

                        for (l = 0; l < n; l++)
                        {
                            if ((slack[l] != 0) && (slack[l] < s))
                            {
                                s = slack[l];
                            }
                        }

                        for (q = 0; q < t; q++)
                        {
                            rowDec[unchosenRow[q]] += s;
                        }

                        for (l = 0; l < n; l++)
                        {
                            if (slack[l] != 0)
                            {
                                slack[l] -= s;

                                if (slack[l] == 0)
                                {
                                    // Begin look at a new zero 22
                                    k = slackRow[l];
                                    if (rowMate[l] < 0)
                                    {
                                        for (j = l + 1; j < n; j++)
                                        {
                                            if (slack[j] == 0)
                                            {
                                                colInc[j] += s;
                                            }
                                        }

                                        breakthru = true;

                                        break; //for-loop
                                    }
                                    else
                                    {
                                        parentRow[l] = k;
                                        unchosenRow[t++] = rowMate[l];
                                    }

                                    // End look at a new zero 22
                                }
                            }
                            else
                            {
                                colInc[l] += s;
                            }
                        }

                        // End introduce a new zero into the matrix 21
                    }
                }

                while (true)
                {
                    j = colMate[k];
                    colMate[k] = l;
                    rowMate[l] = k;

                    if (j < 0)
                    {
                        break;
                    }

                    k = parentRow[j];
                    l = j;
                }

                // End update the matching 20
                if (--unmatched == 0)
                {
                    break;
                }

                // Begin get ready for another stage 17
                t = 0;

                for (l = 0; l < n; l++)
                {
                    parentRow[l] = -1;
                    slack[l] = Double.POSITIVE_INFINITY;
                }

                for (k = 0; k < m; k++)
                {
                    if (colMate[k] < 0)
                    {
                        unchosenRow[t++] = k;
                    }
                }

                // End get ready for another stage 17
            }
        }

        // Begin doublecheck the solution 23
        //   for (k=0;k<m;k++)
        //     for (l=0;l<n;l++)
        //       if (p->cost[k][l]<row_dec[k]-col_inc[l])
        // 	exit(0);
        //   for (k=0;k<m;k++)
        //     {
        //       l=col_mate[k];
        //       if (l<0 || p->cost[k][l]!=row_dec[k]-col_inc[l])
        // 	exit(0);
        //     }
        //   k=0;
        //   for (l=0;l<n;l++)
        //     if (col_inc[l])
        //       k++;
        //   if (k>m)
        //     exit(0);
        // End doublecheck the solution 23
        // End Hungarian algorithm 18
        for (i = 0; i < m; ++i)
        {
            p.assignment[i][colMate[i]] = HUNGARIAN_ASSIGNED;
            p.zuordnung[0][i] = i;
            p.zuordnung[1][i] = colMate[i];
        }

        for (k = 0; k < m; ++k)
        {
            for (l = 0; l < n; ++l)
            {
                p.cost[k][l] = p.cost[k][l] - rowDec[k] + colInc[l];
            }
        }

        for (i = 0; i < m; i++)
        {
            cost += rowDec[i];
        }

        for (i = 0; i < n; i++)
        {
            cost -= colInc[i];
        }

        return cost;
    }
}
///////////////////////////////////////////////////////////////////////////////
//  END OF FILE.
///////////////////////////////////////////////////////////////////////////////
